/* 
	HexEditApplication.cpp

	Author:			Max Horn
	Description:	<describe the HexEditApplication class here>
*/

#include	"HexEditApplication.h"
#include	"HexGotoDialog.h"
#include	"HexSearchDialog.h"
#include	"HexEditView.h"
#include	"PrefsDialog.h"
#include	"MacZoop.h"
#include	"ZPrefsFile.h"
#include	"ZViewDialog.h"

#include	"ZScrollingAboutView.h"
#include	"ZTabSelectorView.h"
#include	"ZURLTextBoxView.h"


HexSearchDialog*	gSearchDialog = 0;
HexGotoDialog*		gGotoDialog = 0;
PrefsDialog*		gPrefsDialog = 0;


HexEditApplication::HexEditApplication()
	: ZApplication()
{
}


HexEditApplication::~HexEditApplication()
{
}


void	HexEditApplication::StartUp()
{
	gPrefsFile = new ZPrefsFile();
	
	if ( gPrefsFile )
		gPrefsFile->OpenResFork();

	gMenuBar->NominateWindowsMenu( 134 );

	//
	// Init InternetConfig
	//
  	OSStatus icerr = ICStart(&mICInstance, gAppSignature);
	if (icerr == noErr)
	{
		icerr = ICFindConfigFile(mICInstance, 0, nil);
 	}
 	else
 	{
 		mICInstance = nil;
 	}
	ZURLTextBoxView::Init(mICInstance);
	
	//
	// Create Goto Dialog
	//
	FailNIL( gGotoDialog = new HexGotoDialog());
	try
	{
		gGotoDialog->InitZWindow();
		gGotoDialog->Place();
	}
	catch( ZoopError err )
	{
		ForgetObject( gGotoDialog );
		
		throw err;
	}
	
	//
	// Create Search Dialog
	//
	FailNIL( gSearchDialog = new HexSearchDialog());
	try
	{
		gSearchDialog->InitZWindow();
		gSearchDialog->Place();
	}
	catch( ZoopError err )
	{
		ForgetObject( gSearchDialog );
		
		throw err;
	}

	//
	// Create Prefs Dialog
	//
	FailNIL( gPrefsDialog = new PrefsDialog());
	try
	{
		gPrefsDialog->InitZWindow();
		gPrefsDialog->Place();
	}
	catch( ZoopError err )
	{
		ForgetObject( gPrefsDialog );
		
		throw err;
	}
	
	
	// Create a hex window
	OpenNewWindowType('BINA');
	
}


void	HexEditApplication::ShutDown()
{
	if ( gPrefsFile )
		ForgetObject( gPrefsFile );

	if (mICInstance != NULL)	
		OSStatus icerr = ICStop(mICInstance);
}



void	HexEditApplication::RegisterClasses()
{
	ZApplication::RegisterClasses();
	ZViewDialog::RegisterDialogItem( DICONSTRUCTORFUNCTION( ZScrollingAboutView ), CLASS_ZScrollingAboutView, kInitItemWithFirstParam );
	ZViewDialog::RegisterDialogItem( DICONSTRUCTORFUNCTION( ZURLTextBoxView ), kMagicStringURLTextBoxView, kInitItemWithFirstParam );

/*	ZViewDialog::RegisterDialogItem( CreateArrayBrowserForDialog, kMagicStringListbox, kInitItemWithMagicParams );
	ZViewDialog::RegisterDialogItem( CreateIconArrayBrowserForDialog, kMagicStringIconList, kInitItemWithMagicParams );
	ZViewDialog::RegisterDialogItem( CreateArrayBrowserFromResource, kMagicStringArrayBrowser, kInitItemWithMagicParams );
	ZViewDialog::RegisterDialogItem( DICONSTRUCTORFUNCTION( ZColourMenuView ), kMagicStringColourMenu, kInitItemWithFirstParam );
	ZViewDialog::RegisterDialogItem( CreateNewDigitalView, kMagicStringCounterDisplay, kInitItemWithFirstParam );
	ZViewDialog::RegisterDialogItem( CreateNewDigitalClock, kMagicStringClockDisplay, kInitItemWithFirstParam );
*/
	ZViewDialog::RegisterDialogItem( CreateNewTabSelectorView, kMagicStringTabSelector, kInitItemWithMagicParams );
/*	ZViewDialog::RegisterDialogItem( CreateNewCalendarView, kMagicStringCalendarView, kInitItemWithMagicParams );
	ZViewDialog::RegisterDialogItem( CreateNewRulerView, kMagicStringRulerView, kInitItemWithMagicParams );
*/
}


ZWindow*		HexEditApplication::MakeNewWindowType( OSType aType )
{
	ZWindow* 		aWindow;
	MakeWindowInfo	mwi;
	
	// allow modifier a chance to handle this:
	
	mwi.fType = aType;
	mwi.theWindow = NULL;
	
	FailNIL( aWindow = new ZWindow( this,  128 ));

	try
	{
		HexEditView*		hv;
		HexEditViewHeader*	hh;

		aWindow->InitZWindow();

		FailNIL( hv = new HexEditView( aWindow ));
		
		hv->FitToParent();
		hv->SetAutoSizing( AUTOSIZE( NONE, NONE, FIXEDBOTTOM, FIXEDRIGHT ));
		hv->ChangeSize( 16, 0, 0, 0 );
		hv->SetParentSizeRect();
		

		FailNIL( hh = new HexEditViewHeader( aWindow, NULL, hv ));
		hh->FitToParent(TRUE,FALSE);
		hh->ChangeSize( -1, -1, 16, 1 );
		hh->SetAutoSizing( AUTOSIZE( NONE, NONE, NONE, FIXEDRIGHT ));
		
		
	}
	catch( ZoopError err )
	{
		ForgetObject( aWindow );
		
		throw err;
	}

	return aWindow;
}


/*----------------------------------***  AboutBox  ***----------------------------------*/
/*
access:			public
overrides:		
description: 	handle the "About..." command
ins: 			none
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditApplication::AboutBox()
{
	Str255		versStr = "\p(unknown version)";
	VersRecHndl	vr;
	
	if( ( vr = ( VersRecHndl ) GetResource( 'vers', 1 ) ) != NULL )
	{
		HLock( (Handle) vr );
		CopyPString( (StringPtr) ( ( ( unsigned long ) &( ** vr ).shortVersion[0] ) ), versStr );
		ReleaseResource( (Handle)vr );
	}

	ParamText( versStr, NULL, NULL, NULL );
	
	OpenDialog( kAboutBoxID );
}


void	HexEditApplication::DoPreferences()
{
	if( 0 != gPrefsDialog )
		gPrefsDialog->Select();
}

/*-------------------------------***  HandleCommand  ***--------------------------------*/
/*
access:			public
overrides:		ZApplication
description: 	handle commands for the app as a whole
ins: 			<aCmd> the command to process
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditApplication::HandleCommand( const ZoopCommand aCmd )
{
	switch( aCmd )
	{
		case 400:
			if( 0 != gSearchDialog )
				gSearchDialog->Select();
			break;
		
		case 405:
			if( 0 != gGotoDialog )
				gGotoDialog->Select();
			break;
		
		default:
			ZApplication::HandleCommand( aCmd );
	}
}


/*---------------------------------***  UpdateMenus  ***--------------------------------*/
/*
access:			public
overrides:		ZApplication
description: 	enable menus that apply to the current app state and context
ins: 			none
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditApplication::UpdateMenus()
{
	gMenuBar->EnableCommand( 400 );
	gMenuBar->EnableCommand( 405 );

	ZApplication::UpdateMenus();
}
