/* 
	HexEditView.cpp

	Author:			Max Horn
	Description:	<describe the HexEditView class here>
*/

#include	"MacZoop.h"
#include	"HexEditView.h"
#include	"HexEditEngine.h"
#include	"ZGrafState.h"
#include	"ZWindow.h"

#include	<stdio.h>
#include	<string.h>


// global line cache speeds up hex dump drawing

static char		lineCache[ ( BYTES_PER_LINE * 3 ) + ( BYTES_PER_LINE / COLUMN_GROUP_SIZE ) + 1 ];
static char* 	gHex = "0123456789ABCDEF";

/*--------------------------------***  CONSTRUCTOR  ***----------------------------------*/

HexEditView::HexEditView( ZView* aParent, Rect* aFrame )
	: ZScrollView( aParent, aFrame, FALSE, TRUE, TRUE )
{
	FailNIL( fEngine = new HexEditEngine() );
	fSelStart = fSelEnd = 0;
	FailNIL( fSelectionRgn = NewRgn());
	fAsciiEditMode = FALSE;
/////
// begin FIX ME
	fSelStart = 60;
	fSelEnd = 100;
// end FIX ME
/////

	fCaretPhase = FALSE;
	SetCanBeHandler( TRUE );

	SetFontInfo( kFontIDMonaco, 9, 0 );
//	CalcSizes();
//	CalcBounds();
}


/*--------------------------------***  DESTRUCTOR  ***----------------------------------*/

HexEditView::~HexEditView()
{
	if ( fSelectionRgn )	
		DisposeRgn( fSelectionRgn );

	ForgetObject( fEngine );
}


/*---------------------------------***  DrawAll  ***------------------------------------*/
/*
access:			public
overrides:		
description: 	draw the frame. This includes the attached scrollbars
ins: 			none
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/
/*
void	HexEditView::DrawAll()
{
	if ( IsVisible())
	{
		Rect	cr;
		
		Focus();
		PrepareForDrawing();

		GetFrame( &cr );
		OffsetRect( &cr, -cr.left, -cr.top );
		InsetRect(&cr, -1, -1);
		cr.bottom = cr.top + fHeaderHeight + 1;

	#if APPEARANCE_MGR_AWARE
		if ( gMacInfo.hasAppearanceMgr )
		{
			DrawThemeWindowHeader( &cr, ( IsActive() && IsEnabled())? kThemeStateActive : kThemeStateInactive );
		}
		else
		{
	#endif
			RGBBackColor( IsActive()? &gMediumGray: &gVeryLightGray );
			EraseRect( &cr );
			FrameRect( &cr );
	#if APPEARANCE_MGR_AWARE
		}
	#endif

		DrawHeader();
		
		ZScrollView::DrawAll();
	}
}

void	HexEditView::DrawHeader()
{
	Str255		pStr;
	
	MoveTo( 4, fAscent + 2 );
	
	NumToString( fEngine->GetLength(), pStr );
	DrawString( "\pLen: " );
	DrawString( pStr );
	
	DrawString( "\p | Selected: " );
	DrawLongAsHex( fSelStart );
	DrawString( "\p:" );
	DrawLongAsHex( fSelEnd );
}
*/


/*---------------------------------***  DrawContent  ***--------------------------------*/
/*
access:			public
overrides:		
description: 	draw the content of this view
ins: 			none
outs: 			none
notes:			this is heavily based on the code in ZHexEditView
----------------------------------------------------------------------------------------*/

void	HexEditView::DrawContent()
{

	register short	lineStart;
	register long	loop, excess, hs, he, dl;
	register long	base;
	Boolean	 		bSup;
	long			dumpLength;
    RgnHandle   	temp;
	Rect	    	r;
	
//	long			top, bottom;

	dumpLength = fEngine->GetLength();
	TextMode( srcCopy );

	// we are only going to draw what's visible, so get the current clip rect to find out
	// what that is. This is essential if you want the view to print properly.
    
	FailNIL( temp = NewRgn());
	GetClip( temp );
	r = *REGIONBOUNDINGBOX( temp );
	DisposeRgn( temp );

	// start on a whole line boundary:
//	r.top += 2;
	r.top -= r.top % fLineHeight;
//	top = r.top + GetVerticalOffset() * fLineHeight;
	lineStart = r.top + fAscent + 1;
//	bottom = r.bottom + GetVerticalOffset() * fLineHeight;
	
	// compute the offset of the visible area start and end
	
	hs = GetOffsetAtPoint( fHdStart, r.top );
	
	bSup = ( hs > dumpLength );
	
	// make sure drawn area is constrained to the dump limits
	
	hs = MAX( 0, MIN( dumpLength, hs ));
	he = MAX( 0, MIN( dumpLength, GetOffsetAtPoint( fHdStart, r.bottom ) + BYTES_PER_LINE ));

	// figure looping parameters:
	
	dl = he - hs;
	
	excess = dl % BYTES_PER_LINE;
	loop = dl / BYTES_PER_LINE;
	base = hs;
	
	
	// hex is dumped 16 bytes at a time (one line) so we loop through <dumpLength / 16>
	// draw the bulk of the dump:
	
	while( loop-- )
	{
		MoveTo( 6, lineStart );
		DrawHexLine( base );
		
		lineStart += fLineHeight;
		base += BYTES_PER_LINE;
	}
	
	// if there's any data in the trailing line, draw that as well:
	
	MoveTo( 6, lineStart );
	
	if ( excess )
	{
		DrawHexLine( base, excess );
		lineStart += fLineHeight;
	}
	
	// draw the hilite region
	
	if ( ! IsPrinting())
	{
		SetHiliteMode();
		InvertRgn( fSelectionRgn );
		
		// if not active, knock out the middle of the seleciton
		
		if (! IsActive())
		{
			CopyRgn( fSelectionRgn, gUtilRgn );
			InsetRgn( gUtilRgn, 1, 1 );
			SetHiliteMode();
			InvertRgn( gUtilRgn );
		}
	}
	// clear the remainder of the window if we drew the last line

	if ( he == dumpLength )
	{
		Rect   	cr;
		
		GetBounds( &cr );
		cr.top = lineStart - fAscent;
		EraseRect( &cr );
		
		// if last line is completely filled, draw the starting address of the next line
	
		if (( excess == 0 ) && ! bSup )
		{
		#if HD_USE_COLOUR
			RGBForeColor( &gDarkRed );
		#endif
		
			DrawLongAsHex( base );
			DrawChar( ':' );
		}
	}
}


/*---------------------------------***  DrawHexLine  ***--------------------------------*/
/*
access:			protected
overrides:		
description: 	draw a complete line of hex data at the current pen location
ins: 			<lineStartAddr> offset into data buffer where line begins
				<bytesInLine> number of bytes to dump per line
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditView::DrawHexLine( register long lineStartAddr, register char bytesInLine )
{
	// draw a hex dump line 16 bytes long, starting at the current pen position
	
	register long 			p = lineStartAddr;
	register unsigned char	c, pc;
	register unsigned char	i = 0, t = 0;
	
	// first draw the line address offset.
#if HD_USE_COLOUR
	RGBForeColor( &gDarkRed );
#endif
	
	DrawLongAsHex( lineStartAddr );
	DrawChar( ':' );

	// move to the horizontal position <fHdStart>
	
	MoveTo( fHdStart, PORTPENLOCATION( CURRENTPORT )->v );
	
	// we build a cache of the line to be drawn since DrawString is much faster than
	// repeated calls to DrawChar.
	
	lineCache[0] = BYTES_PER_LINE * 3 + ( BYTES_PER_LINE / COLUMN_GROUP_SIZE );
	
	do
	{
		// convert byte pointed to by p to hex display.
		
		if ( i < bytesInLine )
		{
//			pc = *p++;
			pc = fEngine->GetByte(p++);	// FIX ME - optimize
			lineCache[ ++t ] = gHex[( pc & 0xF0 ) >> 4];
			lineCache[ ++t ] = gHex[ pc & 0x0F ];
		}
		else
		{
			lineCache[ ++t ] = ' ';
			lineCache[ ++t ] = ' ';
		}
		lineCache[ ++t ] = ' ';
		
		// add extra space between grouped columns
		
		if ((( i + 1 ) % COLUMN_GROUP_SIZE ) == 0 )
			lineCache[ ++t ] = ' ';
	}
	while( ++i < BYTES_PER_LINE );
	
#if HD_USE_COLOUR
	RGBForeColor( &gDarkBlue );
#endif

	DrawString((ConstStr255Param) lineCache );
	
	// dump line as ascii text (unprintable characters drawn as a dot)
	
	MoveTo( fHdEnd + ASCII_OFFSET, PORTPENLOCATION( CURRENTPORT )->v );

	i = t = 0;
	lineCache[0] = BYTES_PER_LINE;
	p = lineStartAddr;
	
	do
	{
		if ( i < bytesInLine )
		{
//			c = *p++;
			c = fEngine->GetByte(p++);	// FIX ME - optimize
			lineCache[ ++t ] = ( c > 0xD9 || c < 0x20 )? '.' : c; 
		}
		else
			lineCache[ ++t ] = ' ';
	}
	while( ++i < BYTES_PER_LINE );
	
#if HD_USE_COLOUR
	RGBForeColor( &gDarkGreen );
#endif
	
	DrawString((ConstStr255Param) lineCache );
}


/*----------------------------------***  DrawCaret  ***---------------------------------*/
/*
access:			protected
overrides:		
description: 	draw the caret
ins: 			<state> TRUE to draw, FALSE to erase
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditView::DrawCaret( Boolean state )
{
	short	firstVisibleOffset, offset;
	Rect	r;
	
	if ( state == fCaretPhase )
		return;
	
	fCaretPhase = state;

	firstVisibleOffset = GetVerticalOffset() * BYTES_PER_LINE;
	
	if( firstVisibleOffset <= fSelStart )
	{
		offset = fSelStart - firstVisibleOffset;
		
		r.top = ( offset / BYTES_PER_LINE ) * fLineHeight;
		r.bottom = r.top + fLineHeight + 2;
		r.right = CalcCaretHOffset( offset );
		r.left = r.right - CARET_WIDTH;

//		if ( fHalfOff )
//			OffsetRect( &r, fEmSpace, 0 );
			
		PenNormal();		 
		if ( state )
			PenMode( patCopy );
		else
			PenMode( patBic );
		PaintRect( &r );
	}	
}


#pragma mark -


/*---------------------------------***  Find1View  ***----------------------------------*/
/*
EEEEEEKKKK


We use an unusual "bounds" value range - instead of pixels, we use lines! erks


----------------------------------------------------------------------------------------*/

ZView*		HexEditView::Find1View( const Point mouse )
{
	ZView*		fv = NULL;
	Rect		br;
	
	// the point passed is relative to our bounds rect
	
	GetContentRect( &br );
	
	if ( PtInRect( mouse, &br ) && IsVisible() && AcceptsClicks())
	{
		// within the bounds, what about a child?
		
		fv = this;
	}
	
	return fv;
}


/*--------------------------------***  ClickContent  ***--------------------------------*/
/*
access:			public
overrides:		
description: 	respond to a click in the view
ins: 			<mouse> the mouse coordinates local to view
				<modifiers> modifier flags
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditView::ClickContent( const Point mouse, const short modifiers )
{
	// click and drag in the editor. This sets up the selection.
	
	Point		pt, oldMouse = { -1, -1 };
	long		hs, he, dumpLength;
	RgnHandle	selRgn;
	short		dblClick = gApplication->GetClicks();
	Boolean		shiftClick = ( modifiers & shiftKey ) == shiftKey;
	long		offsetStart = MIN( fSelStart, fSelEnd ),
				offsetEnd = MAX( fSelStart, fSelEnd );
	
	FailNIL( selRgn = NewRgn());
	
	DrawCaret( FALSE );
//	fNibPhase = FALSE;
//	fHalfOff = FALSE;
	dumpLength = fEngine->GetLength();
	
	// If shift-click, extend selection. If double-click, select a byte, triple selects a line,
	// four clicks selects the whole lot.
	
	he = MIN( dumpLength, MAX( 0, GetOffsetAtPoint( mouse.h, mouse.v )));
	
	if ( shiftClick )
	{
		if ( he < offsetStart )
			hs = offsetEnd;
		else
			hs = offsetStart;
	}
	else
		hs = he;
		
	if ( dblClick > 1 )
	{
		if ( dblClick == 2 )
			he = hs + 1;
		else
		{
			if ( dblClick == 3 )
			{
				hs -= hs % BYTES_PER_LINE;
				he = hs + BYTES_PER_LINE;
			}
			else
			{	
				hs = 0;
				he = dumpLength;
			}
		}
//		CalcSelection( hs, he, selRgn );
//		SetSelectionRgn( selRgn );
		SetSelectionRange( hs, he );
	}
	
	if ( !shiftClick )
		oldMouse = pt = mouse;
	
	// wait for some movement of the mouse before continuing if it's still down.
	
	if ( !shiftClick && StillDown() && !DeltaPoint( pt, oldMouse ))
	{
		GetMouse( &pt );
	}
	
	// track the mouse drag, setting the selection accordingly
	
	if ( dblClick < 4 )
	{
		do
		{
			GetMouse( &pt );
			
			// depending on whether a double or triple-click, modify end point
			
			if ( dblClick == 3 )
				pt.h = fHdEnd + fEmSpace + fEmSpace;
			
			if ( DeltaPoint( pt, oldMouse ))
			{
				he = MAX( 0, MIN( dumpLength, GetOffsetAtPoint( pt.h, pt.v )));
				
				//CalcSelection( hs, he, selRgn );
				//SetSelectionRgn( selRgn );
				SetSelectionRange( hs, he );
			
				oldMouse = pt;
				
				AutoScroll( pt );
			}
		}
		while( WaitMouseUp());
		
		if ( hs == he )
		{
			SetEmptyRgn( selRgn );
			SetSelectionRgn( selRgn );
		}
	}
	DisposeRgn( selRgn );
	
	SetSelectionRange(hs, he);
}


#pragma mark -


/*----------------------------------***  Activate  ***----------------------------------*/
/*
access:			public
overrides:		
description: 	view becoming active - make selection solid
ins: 			none
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditView::Activate()
{
	ZScrollView::Activate();

	Focus();
	PrepareForDrawing();
	FocusForContent();
	
	UpdateSelectionRgn();
}


/*---------------------------------***  Deactivate  ***---------------------------------*/
/*
access:			public
overrides:		
description: 	view becoming inactive - make selection outlined
ins: 			none
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditView::Deactivate()
{
	ZScrollView::Deactivate();
	
	Focus();
	PrepareForDrawing();
	FocusForContent();
	
	UpdateSelectionRgn();
	
	DrawCaret( FALSE );
}


/*------------------------------------***  Idle  ***------------------------------------*/
/*
access:			public
overrides:		
description: 	blink caret in view
ins: 			none
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditView::Idle()
{
	static long		blinkTime = 0;
	long			bkt;
	
	bkt = TickCount();
	
	if ( bkt > blinkTime + GetCaretTime())
	{
		blinkTime = bkt;
		
		if ( fSelStart == fSelEnd && IsActive())
		{
			// blink the "caret"
			Focus();
			PrepareForDrawing();
			FocusForContent();
			DrawCaret( !fCaretPhase );	
		}
	}
}


/*-----------------------------***  ClickInSamePlace  ***-------------------------------*/
/*
access:			public
overrides:		
description: 	detect double click
ins: 			<a,b> successive click locations
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

Boolean	HexEditView::ClickInSamePlace( const Point a, const Point b )
{
	long	oa, ob;
	
	oa = GetOffsetAtPoint( a.h, a.v );
	ob = GetOffsetAtPoint( b.h, b.v );
	
	return ( oa == ob );
}


/*-------------------------------***  AdjustCursor  ***---------------------------------*/
/*
access:			public
overrides:		
description: 	set the cursor according to the view
ins: 			<mouse> the point relative to the frame
				<modifiers> the modifier flags
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/
/*
void		HexEditView::AdjustCursor( const Point mouse, const short modifiers )
{
	Rect	cr;
	
	GetContentRect( &cr );
	
	if ( PtInRect( mouse, &cr ))
	{
		if ( ShouldDrag( mouse, modifiers ))
			SetCursorShape( DRAG_CURSOR );
		else
			SetCursorShape( iBeamCursor );
	}
	else
		SetCursorShape( ARROW_CURSOR );
}
*/

/*-------------------------------***  SetViewCursor  ***--------------------------------*/
/*
access:			public
overrides:		
description: 	set cursor to i-beam in this view
ins: 			<mouse> mouse point
				<modifiers> modifier key state
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditView::SetViewCursor( const Point mouse, const short modifiers )
{
	if ( ShouldDrag( mouse, modifiers ))
		SetCursorShape( DRAG_CURSOR );
	else
		SetCursorShape( iBeamCursor );
}



#pragma mark -


/*------------------------------***  CALCCARETOFFSET  ***-------------------------------*/
/*	
return pixel offset of byte in a row passed. Used for many calculations, such as caret
positioning and selection.
----------------------------------------------------------------------------------------*/

short	HexEditView::CalcCaretHOffset( long offset )
{
	if ( fAsciiEditMode == TRUE )
		return CalcAsciiHOffset( offset );
	else
		return CalcHexHOffset( offset );
}


short	HexEditView::CalcAsciiHOffset( long offset )
{
	return fHdEnd + ASCII_OFFSET + ( ( offset % BYTES_PER_LINE ) * fEmSpace );
}

short	HexEditView::CalcHexHOffset( long offset )
{
	return MIN( fHdEnd, fHdStart +
						   (( offset % BYTES_PER_LINE ) * fEmSpace * 3 ) +
						   (( offset % BYTES_PER_LINE ) / COLUMN_GROUP_SIZE * fEmSpace ));
}


#pragma mark -


/*-------------------------------***  CALCSELECTION  ***--------------------------------*/
/*	
return in <aRgn> the selection region corresponding to the offset start and end passed.
----------------------------------------------------------------------------------------*/

void	HexEditView::CalcSelection( long offsetStart, long offsetEnd, RgnHandle aRgn )
{
	RgnHandle	rangeTemp, hexTemp, asciiTemp;
	short		firstVisibleOffset;
	Boolean		moveUpHack = FALSE;
	
	if ( offsetEnd < offsetStart )
	{
		long swap = offsetStart;
		offsetStart = offsetEnd;
		offsetEnd = swap;
	}

	firstVisibleOffset = GetVerticalOffset() * BYTES_PER_LINE;
	offsetStart -= firstVisibleOffset;
	offsetEnd -= firstVisibleOffset;
	if( offsetStart < 0 )
	{
		offsetStart = MAX(0,offsetStart+BYTES_PER_LINE);
		offsetEnd = MAX(0,offsetEnd+BYTES_PER_LINE);
		moveUpHack = TRUE;
	}

	if ( offsetEnd == offsetStart )
	{
		SetEmptyRgn( aRgn );
		return;
	}

	// selection consists of a region made up from three rectangles- first line, central
	// block and last line. Some of these may be empty.
	
	FailNIL( rangeTemp = NewRgn());
	FailNIL( hexTemp = NewRgn());
	FailNIL( asciiTemp = NewRgn());
	
	CalcHexSelection( offsetStart, offsetEnd, hexTemp );
	CalcAsciiSelection( offsetStart, offsetEnd, asciiTemp );
	
	UnionRgn( hexTemp, asciiTemp, aRgn );
	if ( IsActive() )
	{
		if ( fAsciiEditMode )
		{
			InsetRgn( hexTemp, 2, 2 );
			DiffRgn( aRgn, hexTemp, aRgn );
		}
		else
		{
			InsetRgn( asciiTemp, 2, 2 );
			DiffRgn( aRgn, asciiTemp, aRgn );
		}
	}
	else
	{
		InsetRgn( hexTemp, 1, 1 );
		InsetRgn( asciiTemp, 1, 1 );
		DiffRgn( aRgn, hexTemp, aRgn );
		DiffRgn( aRgn, asciiTemp, aRgn );
	}
	
	CalcRangeSelection( offsetStart, offsetEnd, rangeTemp );
	UnionRgn( rangeTemp, aRgn, aRgn );
	
	DisposeRgn( asciiTemp );
	DisposeRgn( hexTemp );
	DisposeRgn( rangeTemp );
	
	if( moveUpHack )
	{
		OffsetRgn( aRgn, 0, -fLineHeight );
	}
}

/*-----------------------------***  CALCHEXSELECTION  ***-------------------------------*/
/*	
return in <aRgn> the selection region in the hex area.
----------------------------------------------------------------------------------------*/
void	HexEditView::CalcHexSelection( long offsetStart, long offsetEnd, RgnHandle aRgn )
{
	Rect		a, b, c;
	RgnHandle	temp;
	short		sTop, eTop;
	
	if ( offsetEnd == offsetStart )
	{
		SetEmptyRgn( aRgn );
		return;
	}

	// selection consists of a region made up from three rectangles- first line, central
	// block and last line. Some of these may be empty.
	
	FailNIL( temp = NewRgn());
	
	sTop = ( offsetStart / BYTES_PER_LINE ) * fLineHeight;
	eTop = ( offsetEnd / BYTES_PER_LINE ) * fLineHeight;

	SetRect( &a, CalcHexHOffset( offsetStart ),
				 sTop,
				 fHdEnd,
				 sTop + fLineHeight );
	
	SetRect( &c, fHdStart,
				 eTop,
				 CalcHexHOffset( offsetEnd ),
				 eTop + fLineHeight );
	
	SetRect( &b, c.left, a.bottom, a.right, c.top );
	
	// if the selection is within a single line, build it from a and c by forcing a
	// empty and making c the whole selection:
	
	if ( sTop == eTop )
		a.right = c.left = a.left;
	
	RectRgn( aRgn, &b );
	RectRgn( temp, &a );
	UnionRgn( temp, aRgn, aRgn );
	RectRgn( temp, &c );
	UnionRgn( temp, aRgn, aRgn );
	
	DisposeRgn( temp );
}


/*----------------------------***  CALCASCIISELECTION  ***------------------------------*/
/*	
return in <aRgn> the selection region in the ascii area.
----------------------------------------------------------------------------------------*/
void	HexEditView::CalcAsciiSelection( long offsetStart, long offsetEnd, RgnHandle aRgn )
{
	Rect		a, b, c;
	RgnHandle	temp;
	short		sTop, eTop;
	
	if ( offsetEnd == offsetStart )
	{
		SetEmptyRgn( aRgn );
		return;
	}

	// selection consists of a region made up from three rectangles- first line, central
	// block and last line. Some of these may be empty.
	
	FailNIL( temp = NewRgn());
	
	sTop = ( offsetStart / BYTES_PER_LINE ) * fLineHeight;
	eTop = ( offsetEnd / BYTES_PER_LINE ) * fLineHeight;

	// now compute the same selection within the ascii dump area...
	
	SetRect( &a, CalcAsciiHOffset( offsetStart ),
				 sTop,
				 fHdEnd + ASCII_OFFSET + ( fEmSpace * BYTES_PER_LINE ),
				 sTop + fLineHeight );
	
	SetRect( &c, fHdEnd + ASCII_OFFSET,
				 eTop,
				 CalcAsciiHOffset( offsetEnd ),
				 eTop + fLineHeight );
	
	SetRect( &b, c.left, a.bottom, a.right, c.top );

	// if the selection is within a single line, build it from a and c by forcing a
	// empty and making c the whole selection:
	
	if ( sTop == eTop )
		a.right = c.left = a.left;
	
	RectRgn( aRgn, &b );
	RectRgn( temp, &a );
	UnionRgn( temp, aRgn, aRgn );
	RectRgn( temp, &c );
	UnionRgn( temp, aRgn, aRgn );
	
	DisposeRgn( temp );
}


/*---------------------------****  CALCRANGESELECTION  ***------------------------------*/
/*	
return in <aRgn> the small region on the left representing the range of addresses covered.
----------------------------------------------------------------------------------------*/

void	HexEditView::CalcRangeSelection( long offsetStart, long offsetEnd, RgnHandle aRgn )
{
	Rect		a;
	RgnHandle	temp;
	short		sTop, eTop;

	FailNIL( temp = NewRgn());
	
	sTop = ( offsetStart / BYTES_PER_LINE ) * fLineHeight;
	eTop = ( offsetEnd / BYTES_PER_LINE ) * fLineHeight;

	SetRect( &a, 2, sTop, 7,
				(( offsetEnd % BYTES_PER_LINE ) == 0) ? eTop : eTop + fLineHeight );
	RectRgn( aRgn, &a );
	InsetRect( &a, 2, 2 );
	a.right += 2;
	RectRgn( temp, &a );
	DiffRgn( aRgn, temp, aRgn );

	DisposeRgn( temp );
}


/*------------------------------***  SETSELECTIONRGN  ***-------------------------------*/
/*	
set the selection region to that passed, updating visible region as necessary. Port and
origin must be set prior to calling this.
----------------------------------------------------------------------------------------*/

void	HexEditView::SetSelectionRgn( RgnHandle aRgn )
{
	ASSERT("no fSelectionRgn allocated!", fSelectionRgn != NULL, 0)

	FailNILParam( aRgn );
	
	XorRgn( fSelectionRgn, aRgn, gUtilRgn );
	
	PenNormal();
	SetHiliteMode();
	InvertRgn( gUtilRgn );
	
	CopyRgn( aRgn, fSelectionRgn );
}


void	HexEditView::UpdateSelectionRgn()
{
	RgnHandle tempRgn;
	FailNIL( tempRgn = NewRgn() );

	// recalculate selection region
	CalcSelection( fSelStart, fSelEnd, tempRgn );
	SetSelectionRgn( tempRgn );

	DisposeRgn(tempRgn);
}


#pragma mark -


long	HexEditView::GetLength()
{
	return fEngine->GetLength();
}


void	HexEditView::GetSelectionRange( long &start, long &end )
{
	start = fSelStart;
	end = fSelEnd;
}


void	HexEditView::SetSelectionRange( long start, long end )
{
	fSelStart = start;
	fSelEnd = end;
	UpdateSelectionRgn();
	
	SendMessage( kMsgHexEditViewSelectionChanged, NULL );
}


#pragma mark -


/*---------------------------------***  SetFontInfo  ***--------------------------------*/
/*
access:			public
overrides:		
description: 	set the text characteristics for this view
ins: 			<fontID> the font
				<size> the size of the font
				<style> the style of the font
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditView::SetFontInfo( const short fontID, const short size, const Style style )
{
	ZScrollView::SetFontInfo( fontID, size, style );
	CalcSizes();
	CalcBounds();
	PostContentRefresh();
}


/*-------------------------------***  GetContentRect  ***-------------------------------*/
/*
access:			public
overrides:		
description: 	get interior rectangle (excluding header bar)
ins: 			<aRect> accepts result
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/
/*
void	HexEditView::GetContentRect( Rect* aRect )
{
	ZScrollView::GetContentRect( aRect );
	
	aRect->top += fHeaderHeight;
}
*/

long		HexEditView::GetVerticalOffset()
{
	if ( HasVBar())
	{
		if ( GetControlMaximum( theVBar ) > GetControlMinimum( theVBar ))
			return GetControlValue( theVBar );
	}
	
	return 0;
}


/*---------------------------------*** GetPosition ***----------------------------------*/
/*
access:			public
overrides:		
description: 	return scroll offset - since TextEdit handles this itself, we return 0.
ins: 			<pH, pV> receives offset
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditView::GetPosition( short* pH, short* pV )
{
	*pH = 0;
	*pV = 0;
}


/*---------------------------------***  ScrollTo  ***-----------------------------------*/
/*
access:			public
overrides:		
description: 	scroll the view to the absolute offset passed
ins: 			<hPosition> horizontal positon of frame relative to bounds
				<vPosition> vertical position of frame relative to bounds
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditView::ScrollTo( const short hPosition, const short vPosition )
{
	// HACK: we assume hPosition/vPosition are *relative* not absolute
	// since usually this is called with a value based on GetPosition(),
	// which in turn always returns (0,0)
	short	oldV, newV;
	
	oldV = GetVerticalOffset();
	Focus();
	
	if ( HasVBar())
	{
		if ( GetControlMaximum( theVBar ) > GetControlMinimum( theVBar ))
			SetControlValue( theVBar, oldV + vPosition );
//			SetControlValue( theVBar, vPosition );
	}
	
	newV = GetVerticalOffset();
	ScrollView( 0, oldV - newV );
}


/*----------------------------------***  CalcBounds  ***--------------------------------*/
/*
access:			protected
overrides:		
description: 	calculate the bounds rect
ins: 			none
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void		HexEditView::CalcBounds()
{
	Rect	r;
	
	SetRect( &r, 0, 0,
				fHdEnd + ASCII_OFFSET + 6 + ( BYTES_PER_LINE * fEmSpace ),
				( fEngine->GetLength() / BYTES_PER_LINE ) + 1 );
	
	SetBounds( r );
}


/*-----------------------------------***  CalcSizes  ***--------------------------------*/
/*
access:			protected
overrides:		
description: 	compute the lineheights, etc when the drawing parameters are changed
ins: 			none
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void		HexEditView::CalcSizes()
{
	FontInfo	fi;
	
	Focus();
	PrepareForDrawing();

	GetFontInfo( &fi );
	
	fLineHeight = fi.ascent + fi.descent + fi.leading;
//	fHeaderHeight = fLineHeight + 4;
	fAscent = fi.ascent;
	fEmSpace = fi.widMax;
	fHdStart = ( fEmSpace * 9 ) + 20;
	fHdEnd = fHdStart + ( fEmSpace * (( BYTES_PER_LINE * 3 ) + ( BYTES_PER_LINE / COLUMN_GROUP_SIZE ) - 2 ));

	SetScrollAmount( fEmSpace, 1 );
}


/*----------------------------***  CalcScrollParams  ***--------------------------------*/
/*
access:			protected
overrides:		
description: 	compute control maximums, paging values, etc.
ins: 			none
outs: 			none
notes:			called internally as required
----------------------------------------------------------------------------------------*/

void		HexEditView::CalcScrollParams()
{
	long	excess, val;
	long	visLines;
	Rect	cr;
	
	GetContentRect( &cr );
	visLines = ( cr.bottom - cr.top )/fLineHeight;
	
	if ( HasVBar())
	{
		val = GetControlValue( theVBar );
		
		excess = ( fBounds.bottom - fBounds.top ) - visLines;	
		if ( excess < 0 )
			excess = 0;
		SetControlMaximum( theVBar, excess );
		SetControlMinimum( theVBar, fBounds.top );
		
		if ( excess == 0 )
			SetControlValue( theVBar, 0 );

		if ( GetControlValue( theVBar ) != val )
			PostContentRefresh();
	}
	
	vPage = visLines - vScale;

	// set scrollbar proportion. Note that this function is only available
	// with appearance mgr.
	
#if APPEARANCE_MGR_AWARE
	if ( gMacInfo.hasAppearanceMgr )
	{
		if ( HasVBar())
			SetControlViewSize( theVBar, visLines );
	}
#endif
}



/*-------------------------------***  LocalScroll  ***----------------------------------*/
/*
access:			protected
overrides:		
description: 	actually perform the scroll operation
ins: 			<r> area to scroll
				<dH> h distance to scroll
				<dV> v distance to scroll
				<updateRgn> accumulates revealed area
outs: 			none
notes:			dV is measured in *lines*, not pixels, thus we convert it to pixels
----------------------------------------------------------------------------------------*/

void		HexEditView::LocalScroll( Rect* r, short dH, short dV, RgnHandle updateRgn )
{
	// recalculate selection region
	CalcSelection( fSelStart, fSelEnd, fSelectionRgn );


	::ScrollRect( r, dH, dV * fLineHeight, updateRgn );
}


void		HexEditView::SetParentSizeRect()
{
	Rect		sr;
	ZWindow*	zw = dynamic_cast<ZWindow*>(itsBoss);
	
	if( zw == NULL )
		return;
	
	zw->GetSizeRect( &sr );
	sr.left = sr.right = fHdEnd + ASCII_OFFSET + 6 + ( BYTES_PER_LINE * fEmSpace ) + sbWidth;
	zw->SetSizeRect( sr );
}

/*-------------------------------***  GetOffsetAtPoint  ***-----------------------------*/
/*
access:			protected
overrides:		
description: 	get byte offset of pixel location in bounds given by hLoc, vLoc
ins: 			<hLoc> horizontal position
				<vLoc> vertical position
outs: 			byte offset of position
notes:			
----------------------------------------------------------------------------------------*/

long	HexEditView::GetOffsetAtPoint( short hLoc, short vLoc )
{
	// returns the offset address for the coordinate given. The offset starts at 0 and to
	// compute an absolute address add <dumpArea>. The point passed in could be a mouse
	// location or line offset, depending on the intended use. The horizontal point is
	// treated as being to the right of the displayed byte.
	
	Boolean inAscii = hLoc >= ( fHdEnd + ASCII_OFFSET );
	short	hOffset, hm, lineNum;
	
	vLoc += GetVerticalOffset() * fLineHeight;
	
	lineNum = vLoc / fLineHeight;

	if ( inAscii )
	{
		hLoc = MAX( fHdEnd + ASCII_OFFSET, MIN( fHdEnd + ASCII_OFFSET + ( fEmSpace * BYTES_PER_LINE ), hLoc ));
		hOffset = ( hLoc - ( fHdEnd + ASCII_OFFSET ) + ( fEmSpace / 2 )) / fEmSpace;
	}
	else
	{
		hLoc = MAX( fHdStart, MIN( fHdEnd + ASCII_OFFSET - 1, hLoc ));
		
		// to allow for grouping, hLoc needs to be corrected to allow one emSpace per group.
		// the number of groups to the left of hLoc is given by:
		
		hm = (( hLoc - fHdStart - ( fEmSpace / 2 )) / ( fEmSpace * 3 )) / COLUMN_GROUP_SIZE;
		hLoc -= ( hm * fEmSpace );
		
		hOffset = ( hLoc - fHdStart + fEmSpace + ( fEmSpace / 2 )) / ( fEmSpace * 3 );
	}
	
	return (long) lineNum * BYTES_PER_LINE + hOffset;
}


#pragma mark -


/*----------------------------------***  CONSTRUCTOR  ***-------------------------------*/

HexEditViewHeader::HexEditViewHeader( ZView* aParent, Rect* aFrame, HexEditView* aHexView )
	: ZView( aParent, aFrame )
{
	FailNILParam( fHexView );

	fHexView = aHexView;

	ListenTo( fHexView );
	fHexView->GetSelectionRange( fSelStart, fSelEnd );
	
	SetFontInfo( kFontIDGeneva, 9, 0 );
}


/*----------------------------------***  DESTRUCTOR  ***--------------------------------*/

HexEditViewHeader::~HexEditViewHeader()
{
}



/*---------------------------------***  DrawContent  ***--------------------------------*/
/*
access:			public
overrides:		
description: 	draw the placards that make up the header
ins: 			none
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditViewHeader::DrawContent()
{
	char		buffer[256];
	Rect		cr;
	
	GetContentRect( &cr );
//	InsetRect(&cr, -1, -1);

#if APPEARANCE_MGR_AWARE
	if ( gMacInfo.hasAppearanceMgr )
	{
		DrawThemeWindowHeader( &cr, ( IsActive() && IsEnabled())? kThemeStateActive : kThemeStateInactive );
	}
	else
	{
#endif
		RGBBackColor( IsActive()? &gMediumGray: &gVeryLightGray );
		EraseRect( &cr );
		FrameRect( &cr );
#if APPEARANCE_MGR_AWARE
	}
#endif

	FontInfo	fi;
	GetFontInfo( &fi );
	MoveTo( 4, fi.ascent + 2 );
	
	snprintf(buffer, 256,
			"Len:%6.6lXh | Selected %6.6lX:%6.6lXh",
			(long)fHexView->GetLength(), fSelStart, fSelEnd );
	
	DrawText( buffer, 0, strlen( buffer ) );

/*	NumToString( fHexView->GetLength(), pStr );
	DrawString( "\pLen: " );
	DrawString( pStr );
	
	DrawString( "\p | Selected: " );
	DrawLongAsHex( fSelStart );
	DrawString( "\p:" );
	DrawLongAsHex( fSelEnd );
*/
}


/*-------------------------------***  ReceiveMessage  ***-------------------------------*/
/*
access:			public
overrides:		
description: 	respond to messages from browser
ins: 			<aSender> object sending the message
				<msg> the message
				<msgData> data with message
outs: 			none
notes:			
----------------------------------------------------------------------------------------*/

void	HexEditViewHeader::ReceiveMessage( ZComrade* aSender, ZoopMessage msg, UntypedPtr msgData )
{
	switch ( msg )
	{
		case kMsgHexEditViewSelectionChanged:
			{
				long	start, end;
				fHexView->GetSelectionRange( start, end );
				if( (start != fSelStart) || (end != fSelEnd) )
				{
					fSelStart = start;
					fSelEnd = end;
					Focus();
					PrepareForDrawing();
					DrawContent();
					fHexView->Focus();
				}
			}
			break;
		default:
			break;
	}
}


