/* 
	HexEditView.h

	Author:			Max Horn
	Description:	<describe the HexEditView class here>
*/

#ifndef HexEditView_H
#define HexEditView_H

#include	"ZScrollView.h"



// Hex editor compilation options:
// set this to the number of bytes per line to dump. Usually 16; 8 also common. Other sizes
// should work but not exhaustively tested.

#define		BYTES_PER_LINE				16

// set gap between end of dump area and start of ascii area:

#define		ASCII_OFFSET				20

// group into columns of this many bytes (usually 4 for 16 bytes per line or
// 2 for 8 bytes/line ) set equal to BYTES_PER_LINE for no grouping.

#define		COLUMN_GROUP_SIZE			BYTES_PER_LINE

// caret is normally 1 or 2 pixels wide- set your preference here:

#define		CARET_WIDTH					1


class	HexEditEngine;


class HexEditView: public ZScrollView
{
protected:
	HexEditEngine*	fEngine;

	Boolean			fAsciiEditMode;
	long			fSelStart, fSelEnd;

	RgnHandle		fSelectionRgn;
	short			fHdStart, fHdEnd;
	short			fLineHeight, fAscent, fEmSpace;
//	short			fHeaderHeight;
	Boolean			fCaretPhase;
	
public:
	HexEditView( ZView* aParent, Rect* aFrame = NULL );
	virtual ~HexEditView();
	
	
//	virtual void	DrawAll();
//	virtual void	DrawHeader();
	virtual void	DrawContent();

	virtual void	SetFontInfo( const short fontID, const short size, const Style style );
	virtual void	ClickContent( const Point mouse, const short modifiers );
//	virtual void	Type( const char theKey, const short modifiers );
	virtual void	Activate();
	virtual void	Deactivate();
	virtual void	Idle();
	virtual Boolean	ClickInSamePlace( const Point a, const Point b );
//	virtual void	AdjustCursor( const Point mouse, const short modifiers );
	virtual void	SetViewCursor( const Point mouse, const short modifiers );

	
	virtual void	ScrollTo( const short hPosition, const short vPosition );
	virtual void	GetPosition( short* pH, short* pV );
//	virtual void	GetContentRect( Rect* aRect );

	
	long			GetVerticalOffset();
	
	void			SetParentSizeRect();

	// FIX ME!!!
	ZView*			Find1View( const Point mouse );
	
	
	long			GetLength();
	void			GetSelectionRange( long &start, long &end );
	void			SetSelectionRange( long start, long end );
	

protected:

// rendering:
	inline 	void	DrawHexLine( long lineStartAddr, char bytesInLine = BYTES_PER_LINE );
	void			DrawCaret( Boolean state );
	long			GetOffsetAtPoint( short hLoc, short vLoc );

	short			CalcCaretHOffset( long offset );
	short			CalcAsciiHOffset( long offset );
	short			CalcHexHOffset( long offset );

	void			CalcSelection( long offsetStart, long offsetEnd, RgnHandle aRgn );
	void			CalcHexSelection( long offsetStart, long offsetEnd, RgnHandle aRgn );
	void			CalcAsciiSelection( long offsetStart, long offsetEnd, RgnHandle aRgn );
	void			CalcRangeSelection( long offsetStart, long offsetEnd, RgnHandle aRgn );
	void			SetSelectionRgn( RgnHandle aRgn );
	void			UpdateSelectionRgn();

// overrides
	void			CalcBounds();
	void			CalcSizes();
	virtual void	CalcScrollParams();
	virtual void	LocalScroll( Rect* r, short dH, short dV, RgnHandle updateRgn );

	
};


class HexEditViewHeader: public ZView
{
protected:
	HexEditView*		fHexView;

	long				fSelStart, fSelEnd;
	
public:
	HexEditViewHeader( ZView* aParent, Rect* aFrame, HexEditView* aHexView );
	virtual ~HexEditViewHeader();

	virtual void		DrawContent();
//	virtual void		ClickContent( const Point mouse, const short modifiers );
//	virtual void		Activate();
//	virtual void		Deactivate();
	virtual void		ReceiveMessage( ZComrade* aSender, ZoopMessage msg, UntypedPtr msgData );

};


// messages:

enum
{
	kMsgHexEditViewSelectionChanged 	= 'HVsx'
};


#endif	// HexEditView_H
